/*******************************************************
 * cookie-panel.js
 *
 * Authors: 	Alex Quinn (aquinn@cs.washington.edu)
 *				Daniel Howe (dchowe@u.washington.edu)
 *				Carlos Arguelles (carloarg@cs.washington.edu)
 *
 * Based on a project done at the University Of Washington
 * Information School led by Professor Batya Friedman
 * (batya@u.washington.edu).  The project began around
 * 2000 or thereabouts.
 *
 * Updated 2/13/2006 to support Firefox 1.5
 *
 ******************************************************/


/* 
 * NOTES ABOUT COMPATIBILITY DIFFERENCES BETWEEN FIREFOX 1.0 AND 1.5
 *
 * In Firefox 1.0 and before, there was a bug in the listbox code.  In
 * code such as this which uses a listbox that includes a listhead or
 * listcols element, the insertItemAt and removeItemAt functions
 * did not work properly.  If you specified index 2, you would get the
 * first element (index 0).  Essentially, you had to add 2 to the parameter
 * whenever using those two functions.  I reported this as bug #144708,
 * but it was not fixed due to the developer's inability to reproduce it
 * and my lack of follow-up.  Finally, somebody else opened a new bug
 * #236068 and it got fixed.  My old code for version 1.0 compensated
 * by simply adding 2 to the parameter of these calls.  This compensation
 * broke the extension when Firefox 1.5 was released.  So, I had to add
 * this ugly call to compensate only for version 1.0.  This would be
 * a problem for earlier versions of Firefox and Mozilla as well, but I am
 * not supporting those pre-1.0 versions, so it's not an issue to me.
 * I added this fix on February 12, 2006.
 *
 * Anyway, that's why this strange piece of code is here.
 *
 * Also, the string that I have to use to catch a cookie event (I don't totally
 * understand how that mechanism works.) changed from Firefox 1.0 to 1.5.  In
 * 1.0, I had to set it up with the string "cookieChanged".  When I updated to
 * support Firefox 1.5, I found that string had been deprecated and only
 * cookie-changed would work.  So, based on the userAgent, I use one or the other.
 * There's probably a better way to detect the browser version, but I don't know
 * it.
 */

var bug236068compensation = 0;
var observerString = "cookie-changed";
if( navigator.userAgent.indexOf("Firefox/1.0")>0 ) {
	bug236068compensation = 2;
	observerString = "cookieChanged";
}


// XPCOM constants
const COOKIE_MAN_CID = "@mozilla.org/cookiemanager;1";
const g_nsICookieManager = Components.interfaces.nsICookieManager;
const g_nsICookie = Components.interfaces.nsICookie;

// CSS classes for different types of listbox item
const CSS_CLASS_LABEL = "cookie-label";
const CSS_CLASS_FOREIGN = "cookie-foreign";
const CSS_CLASS_NOTFOREIGN = "cookie-notforeign";
const CSS_CLASS_NEWFOREIGN = "cookie-new-foreign";
const CSS_CLASS_NEWNOTFOREIGN = "cookie-new-notforeign";

// Expiration categories (essentially, an enumeration)
const EXP_TEMPORARY = 0;
const EXP_TODAY = 1;
const EXP_YEAR = 2;
const EXP_NEVER = 3;
const EXP_LABELS = ["Temporary cookies",
					"Cookies exp today",
					"Cookies exp this year",
					"Permanent cookies"]
var CREATION_LABEL_HOUR = "Past hour";
var CREATION_LABEL_DAY = "Past day";
var CREATION_LABEL_WEEK = "Past week";
var CREATION_LABEL_MONTH = "Past month";
var CREATION_LABEL_YEAR = "Past year";
var CREATION_LABEL_FOREVER = "Very old";

// Sorting/grouping modes
const SORT_MODE_EXPIRATION = 0;
const SORT_MODE_DOMAIN = 1;
const SORT_MODE_CREATION = 2;
const SORT_MODE_DEFAULT = SORT_MODE_DOMAIN;
const SORT_LABEL_EXPIRATION = "Sort by Expiration Date...";
const SORT_LABEL_DOMAIN = "Sort by Domain...";
const SORT_LABEL_CREATION = "Sort by Creation Time...";
const SORT_LABEL_DEFAULT = SORT_LABEL_DOMAIN;

// The current state of the sorting in the listbox.
var g_sortMode = SORT_MODE_DEFAULT;

// Chrome URL for the HTML file that gives the user general educational information
// about cookies.
const COOKIEINFO_HTML = "chrome://cookiepanel/locale/cw-about-cookies.html";
const COOKIEINFO = "cookie-info"; // string used to launch the window

// Filename for our own cookie store, which contains a little information about the
// cookies in addition to what is in the built-in cookie store (cookies.txt).
const COOKIEPANEL_FILENAME = "cookiepanel.txt";

// Flags to let us turn off the debugging code and such from one central location
// in the code.
const DISABLE_LOG_FILE = true;  // disables the dmp function
const DISABLE_DEBUG_ALERTS = true;  // disables the admp function
const ENABLE_USER_ALERTS = true; // useful for testing without interruptions

// XPCOM object which gives access to the built-in Mozilla cookie file.
var g_cookiePanel = null;

// After 1 second (1000 milliseconds) from the time the last cookieChanged event
// was triggered, we will update the CookiePanel.  Without this delay, if a site
// set 100 cookies, the panel would be updated 100 times inside of a second or two.
// That would slow down the browser terribly.
var MILLISECONDS_DELAY_TO_REBUILD_PANEL = 1000;

// When new cookies come in, we highlight them and attach a little extra text
// for 5 seconds (5000 milliseconds).
var MILLISECONDS_TO_HIGHLIGHT_NEW_COOKIES = 5000;

// timersSet is the number of outstanding timers that have been set to refresh the panel.
// Here's the problem and the solution.  When browsing the web, sometimes a site will set
// several cookies and perhaps delete some all at once.  Each one triggers an event to
// refresh the panel.  So, my solution is to wait one second (or whatever) until after
// the last event before we refresh the panel.  That way, it will only refresh once.
// Otherwise, it would have some significant performance issues.  Currently, it takes
// about 250 milliseconds to refresh a panel of 98 cookies.  For someone with 1000 cookies,
// that might be more like 3 or 5 seconds.  If a site set 10 cookies, that could leave the
// user sitting there for 30 to 50 seconds.
var timersSet = 0;




/**************************************************************
 *
 * Functions for DEBUGGING
 *
 **************************************************************/

function admp (msg) {
/*
 * For debugging only.
 *
 * Display the specified message in an alert box.  This function exists
 * to make it easy to find these calls in the code and to make it easy to
 * disable them all with one constant flag setting.
 */
	if( DISABLE_DEBUG_ALERTS==false ) {
		alert(msg);
	}
}
function dmpNoNewLine(msg) {
/*
 * For debugging only.
 *
 * Write the specified message to a debugging log file without
 * a trailing new line character.
 */
	if( DISABLE_LOG_FILE==false ) {
		var dmpFile = new File("c:\\ff\\dmp.txt");
		dmpFile.open("a");
		dmpFile.write(msg);
		dmpFile.close();
	}
}
function dmp (msg) {
/*
 * For debugging only.
 *
 * Write the specified message to a debugging log file.
 */
	if( DISABLE_LOG_FILE==false ) {
		var dmpFile = new File("c:\\ff\\dmp.txt");
		dmpFile.open("a");
		dmpFile.write(msg + "\n");
		dmpFile.close();
	}
}
function clearDmpFile() {
/*
 * For debugging only.
 *
 * Clear the debugging log file.  Typically done once at the beginning of the
 * session.
 */
	if( DISABLE_LOG_FILE==false ) {
		var dmpFile = new File("c:\\ff\\dmp.txt");
		dmpFile.open("w");
		dmpFile.write("----------------------------------------------\n");
		dmpFile.write((new Date()).toLocaleString() + "\n");
		dmpFile.write("----------------------------------------------\n");
		dmpFile.close();
	}
}

function printable(cookie) {
/*
 * For debugging only.
 *
 * Return an abbreviated string representation of the cookie which will be easily readable
 * in the debugging log file.
 */
	return "<h=\""+cookie.host+"\"/\"n="+cookie.name+"\"/\"v="+cookie.value+"\">";
}




/**************************************************************
 *
 * Functions for BUILDING THE LISTBOX
 *
 **************************************************************/

function buildPanel() {
/*
 * This is the function to call to build or update the sidebar.  Takes care of
 * everything.
 */
				/*for testing*/  var startTime = (new Date()).getTime();
	var builtInCookiesHash = getBuiltInCookiesHash();
				/*for testing*/  var step1 = (new Date()).getTime();
	var filedCookiesHash = getFiledCookiesHash();
	combineBuiltInCookiesWithFiledCookies(builtInCookiesHash,filedCookiesHash);
				/*for testing*/  var step2 = (new Date()).getTime();
	fillCookiePanel(builtInCookiesHash);
				/*for testing*/  var step3 = (new Date()).getTime();
	saveCookiePanelFile(builtInCookiesHash);
				/*for testing*/  var step4 = (new Date()).getTime();
	setCookiePreferenceWidget();
	var timingMsg = ("buildPanel took " + ((new Date()).getTime()-startTime) + " milliseconds.\n" + /*for testing*/
		"*\tFetching cookies.txt took " + (step1-startTime) + " milliseconds.\n" +
		"*\tFetching cookiepanel.txt took " + (step2-step1) + " milliseconds.\n" +
		"*\tClearing the panel and adding new items took " + (step3-step2) + " milliseconds.\n" +
		"*\tSaving cookiepanel.txt took " + (step4-step3) + " milliseconds.\n" +
		"");
}

function fillCookiePanel(cookiesHash) {
/*
 * Main function for filling the listbox.
 *
 * The speed of this function is very important.  Since this function gets run
 * fairly frequently - at least once for every page loaded that adds cookies - if it
 * is slow, the CookiePanel extension will slow down the browser significantly.
 * This was rewritten in September 2005.  Previously, it was so slow that it took
 * 4 to 30 seconds every time the browser was opened and every time cookies were added.
 * I have no doubt it could be made faster, but it is definitely much faster than before.
 * On my 1.5 GHz 512 MB Pentium M laptop with about 100 cookies, it takes about
 * 300 milliseconds.
 *
 * Here's one way I speeded it up.  Every time a cookie is added or deleted and you
 * want to update the listbox, you basically have to rebuild it from scratch.  At first
 * I thought I could just add only the new cookies or delete only the cookies which were
 * removed.  However, the cookie interface doesn't tell you which cookies were added
 * or removed.  It only tells you that something happened.  You have to figure it out for
 * yourself.  It might be possible to detect what was added or removed except for the
 * fact that some web pages might delete AND add cookies.  So, at first, I cleared the
 * whole panel and then added new listitems for every item in the list.  That was slow.
 * So now I leave the items right where they are.  I simply change the label and
 * CSS class for each item.  If cookies were deleted, then there will be unused items at
 * the end.  We delete them.  If cookies were added, then there won't be enough items in
 * the list at first to display all the cookies.  So, we add new listitems to the end.
 * It makes it a little more complicated, but it's also a lot faster.  (AQ 9/27/05)
 */
	if( hashtableIsEmpty(cookiesHash) ) {
		clearCookiePanel();
	}
	else {
		var cookieArray = new Array();
		var cookie = null;
		var cookieCount = 0;
		var foundNewCookiesFlag = false;
		var firstNewCookiePosition = "";
		var lastNewCookiePosition = "";
		for( var key in cookiesHash ) {
			cookie = cookiesHash[key];
			cookieArray.push( Array(
								constructLabelForCookie(cookie), // 0
								cookie, // 1
								makeCookieItemText(cookie), // 2
								key, // 3
								getCssClassForCookie(cookie))); //4
			cookieCount++;
		}
		setCookieCounter(cookieCount);
		cookieArray.sort(compareCookieTuple);
			
		var cookieListbox = document.getElementById("cookieList");
		var cookieListboxLength = cookieListbox.getRowCount();
		var rowsSoFar = 0;
		var item = null;
		var cookieCount = cookieArray.length;
		var lastHeader = "", headerCount = 0;
		var prevSelectedItem = cookieListbox.selectedIndex;

		if( cookieListboxLength>0 ) {
			item = cookieListbox.getItemAtIndex(0);
		}

		// At this point, item is null if the listbox was empty.  If it was not empty,
		// then it refers to the first item in the list.  This first item in the list
		// will be overwritten with something.

		for(var j=0; j<cookieCount; j++) { // For 100 cookies, i will be in [100...1]
			var position = j + headerCount; // For 100 cookies and 3 headers, position will be in [0...102]

			// At this point, if it's the first iteration, item might be null.  Usually, it
			// points to the cell in which we're about to write.

			// Add a header, if necessary.
			if( cookieArray[j][0] != lastHeader ) {
				// If there are enough rows in the listbox already, then use an existing listitem.
				if(position<cookieListboxLength) {
					item.setAttribute("label",cookieArray[j][0]);
					item.setAttribute("value","");
					item.setAttribute("class",CSS_CLASS_LABEL);
					item.setAttribute("context","");
					if(position+1<cookieListboxLength) {
						item = cookieListbox.getNextItem(item,1);
					}
				}
				// If there are not enough preexisting rows in the listbox, then we add a new listitem
				// to the end.
				else {
					item = cookieListbox.appendItem(cookieArray[j][0]);
					item.setAttribute("value","");
					item.setAttribute("class",CSS_CLASS_LABEL);
					cookieListboxLength++;
				}
				headerCount++;
				position++;
				lastHeader = cookieArray[j][0];
			}

			// Add the cell for the cookie.

			// If there are enough rows in the listbox already, then use an existing listitem.
			if(position<cookieListboxLength) {
				item.setAttribute("label",cookieArray[j][2]);
				item.setAttribute("value",cookieArray[j][3]);
				item.setAttribute("class",cookieArray[j][4]);
				item.setAttribute("context","cookieListMenu");
				item.setAttribute("ondblclick","viewSelectedCookie();");
				if(position+1<cookieListboxLength) {
					item = cookieListbox.getNextItem(item,1);
				}
			}
			// If there are not enough preexisting rows in the listbox, then we add a new listitem
			// to the end.
			else {
				item = cookieListbox.appendItem(cookieArray[j][2]);
				item.setAttribute("value",cookieArray[j][3]);
				item.setAttribute("class",cookieArray[j][4]);
				item.setAttribute("context","cookieListMenu");
				item.setAttribute("ondblclick","viewSelectedCookie();");
				cookieListboxLength++;
			}

			// We keep track of the location of the first (uppermost) and last (lowermost)
			// cookies in the listbox.  Later, we will do some scrolling to make them visible.
			if(cookieIsNew(cookieArray[j][1])) {
				foundNewCookiesFlag = true;
				if( firstNewCookiePosition=="" ) {
					firstNewCookiePosition = position;
				}
				lastNewCookiePosition = position;
			}
		}
		while( cookieListboxLength>cookieCount+headerCount ) {
			cookieListbox.removeItemAt((cookieListboxLength-1)+bug236068compensation);
			cookieListboxLength--;
		}
		if( foundNewCookiesFlag==true ) {
			// For a few seconds, we want any new cookies to be highlighted and have a label
			// attached pointing them out to the user.  That has already been done by the
			// functions that choose the CSS class and the text for the listitem.  But, we want
			// to be sure to remove that information in a few seconds after they're no longer
			// considered new.  So, we set a timer to go back over them in a few seconds.
			setTimeout("reviewNewCookies();",MILLISECONDS_TO_HIGHLIGHT_NEW_COOKIES+100);

			// We want the user to be able to see new cookies that are being set as much as
			// possible.  Otherwise, the CookiePanel doesn't really do its job of making
			// the user aware of the cookies.  If the first cookie in the panel is new, then
			// we cheat a little and make the first heading visible too, mainly aesthetics.
			// It would be complicated to do this perfectly.
			cookieListbox.ensureIndexIsVisible(lastNewCookiePosition);
			if( firstNewCookiePosition==1 ) {
				cookieListbox.ensureIndexIsVisible(0);
			}
			else {
				cookieListbox.ensureIndexIsVisible(firstNewCookiePosition);
			}
		}
		if( cookieListbox.getRowCount()>0 ) {
			cookieListbox.selectedIndex = Math.min(prevSelectedItem,cookieListbox.getRowCount()-1);
		}
	}
}

function clearCookiePanel() {
/*
 * Remove everything from the listbox.  Actually, this function isn't
 * necessary.  The same thing would happen if we just called
 * fillCookiePanel, but I think that would be slower.
 */
	var cookieListbox = document.getElementById("cookieList");
	var cookieListboxLength = cookieListbox.getRowCount();
	var i=0;
	for(i=cookieListboxLength; i!=0; i--) {
		cookieListbox.removeItemAt(0+bug236068compensation);
	}
	setCookieCounter(0);
}




/**************************************************************
 *
 * Functions for HASHTABLE MANIPULATION
 *
 **************************************************************/
function hashtableIsEmpty(hashtable) {
/*
 * When using the JavaScript Object as a hashtable, this returns
 * true iff the hashtable is empty (IOW, the object has no properties).
 */
	for( var o in hashtable ) {
		return false;
	}
	return true;
}

function hashHasKey(hashTable,key) {
/*
 * When using the JavaScript Object as a hashtable, this returns
 * true iff the hashtable (object) has the specified key.
 */
	return typeof(hashTable[key])!="undefined";
}



/**************************************************************
 *
 * Functions for RESPONDING TO EVENTS FROM THE XUL
 *
 **************************************************************/

function displayCookieText() {
/*
 * Display the educational information about cookies.  This information is stored as HTML
 * files right in the chrome hierarchy.
 */
	window.open(COOKIEINFO_HTML,COOKIEINFO,"");

	// The flags were causing problems.  I might go back to them some day.
	// "scrollbars=yes,status=no,menubar=no,toolbar=no,resizeable=yes,width=550,height=400,top=50,left=40");
}

function done() {
/*
 * Just a placeholder right now.  Is called when the CookiePanel is closed.
 */
}

//      In version 0.6 of the CookiePanel, I had a button to open Firefox's
//      built-in cookie manager.  In Firefox 1.5, that cookie manager changed
//      quite a bit, so I scrapped that feature.  This function is shrappnel
//      from that old feature.  I'm leaving it here, for the time being, in case
//      something comes up that allows me to restore that feature.
//
//
//  function launchCookieViewer() {
//  /*
//   * Launch the built-in cookie manager.
//   *
//   * Note that it is not possible to simply call the code for the built-in
//   * cookie manager directly because it interacts with the preferences dialog.
//   * It is set up so that if you delete cookies, they aren't deleted until you
//   * click OK to the PREFERENCES dialog (not the cookie manager).  In other words,
//   * it doesn't delete them when you click OK to the cookie manager.  It simply
//   * passes the info about which cookies to delete down to the preferences dialog
//   * code which takes care of it later.  This is not possible to do from the
//   * CookiePanel, so I had to make a copy of the cookie manager code and include
//   * it in the CookiePanel XPI.  This has the advantage of preventing glitches
//   * in the event that the built in cookie manager changes some time.  Of course,
//   * my copied off cookie manager has been modified to take care of deleting the
//   * cookies itself and such.
//   */
//  	window.openDialog("chrome://cookiepanel/content/cookieviewer/CookieViewer.xul","_blank",
//  					  "chrome,resizable=yes", "cookieManager");
//  }

function init()
{
/*
 * This should run only once, when the sidebar panel is initialized.
 */
	// Make sure jslib is installed.
	jslibCheckInst();

	// Initialize jslib
	jslib.init(this);
	include (jslib_file);
	include (jslib_dirutils);

	// Clear the debugging log file.
	clearDmpFile();

	/* The following two lines were copied almost verbatim from CookieViewer.js
	 * which was written by Ben Goodger.  Thanks also to Dan Whitte for pointing
	 * me in the right direction.  4-1-03 ajq
	 *
	 * These lines serve to set up an observer so my code can be notified
	 * when cookies are set, changed, or deleted.
	 */
	var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
	observerService.addObserver(cookieReloadListener, observerString, false);

	// Initialize an XPCOM object to read from the main cookie store.
	g_cookiePanel = Components.classes[COOKIE_MAN_CID].getService();
	g_cookiePanel = g_cookiePanel.QueryInterface(g_nsICookieManager);

	// Set the listbox header with the desired title.
	var headRow = document.getElementById("headRowCell");
	headRow.setAttribute("label", SORT_LABEL_DEFAULT);
	buildPanel();
}



/**************************************************************
 *
 * Functions for INTERACTING WITH THE COOKIE STORES
 *
 **************************************************************/

function combineBuiltInCookiesWithFiledCookies(builtInCookiesHash,filedCookiesHash) {
/*
 * We keep our own cookie file, cookiepanel.txt, in addition to the built-in 
 * one (cookies.txt).  After we've read them both in and stored them in hashtables,
 * this function combines them.  It takes all the info from the built-in cookie
 * store and adds the extra information from our cookiepanel.txt file.
 */
	var filedCookie = null, cookieKeyName;
	var cookieCount = 0;
	var cookie = null;
	for(var cookieKeyName in builtInCookiesHash) {
		cookie = builtInCookiesHash[cookieKeyName];
		filedCookie = filedCookiesHash[cookieKeyName];
		if( typeof(filedCookie)=="object" ) {
			cookie.isForeign = filedCookie.isForeign;
			cookie.creationDate = filedCookie.creationDate;
		}
		cookieCount++;
	}
	return cookieCount;
}

function getCookiePanelFile() {
/*
 * Get a jsLib file object for the cookiepanel.txt file.  This object
 * will be used by other functions to read/write that file.
 */
	var du = new DirUtils();
	var f = new File(du.getPrefsDir());
	f.append(COOKIEPANEL_FILENAME);
	return f;
}

function getBuiltInCookiesHash() {
/*
 * Get the cookies from the built-in cookie store (cookies.txt) via
 * the XPCOM interface to the cookie store.  Return them in the form
 * of a hash table.
 */
	var enumerator = g_cookiePanel.enumerator;
	var cookieHash = new Object();  // This will be a hashtable.
	var cookie = null;
	var mozillaCookie = null;
	while( enumerator.hasMoreElements() ) {
		mozillaCookie = enumerator.getNext().QueryInterface( g_nsICookie );
		cookie = convertCookie( mozillaCookie );
		cookieHash[makeCookieKeyName(cookie)] = cookie;
	}
	return cookieHash;
}

function getFiledCookiesHash() {
/*
 * Fetch the information from the cookiepanel.txt file.  This file contains
 * everything that's in the main cookies.txt file (built-in cookie store) as
 * well as some other information the the CookiePanel needs but is not in the
 * built in cookie store.  Specifically, it also includes the creation date
 * and whether or not the cookie is a foreign cookie or not.  (Foreign cookie
 * detection is currently disabled, but we still keep a dummy flag, just in case
 * we find a way to make it work in the future.) (AQ 9/27/05)
 */
	var cpf = getCookiePanelFile();
	var cookie = null;
	var cookieHash = new Object();
	var s = "";
	var cookiepanelFileLines = null;
	var cookiepanelFileText = "";
	var thisLine = "";
	if( cpf.exists() ) {
		cpf.open("r");
		cookiepanelFileText = cpf.read();
		if( cookiepanelFileText!="" ) {
			cookiepanelFileLines = cookiepanelFileText.split("\n");
			for( var i=cookiepanelFileLines.length; i!=0; i-- ) {
				thisLine = cookiepanelFileLines[i-1];
				if(thisLine!="") {
					cookie = unserializeCookie(thisLine);
					cookieHash[makeCookieKeyName(cookie)] = cookie;
				}
			}
		}
	}
	return cookieHash;
}

function saveCookiePanelFile(cookiesHash) {
/*
 * Save our own version of the cookie store.  The built-in cookie store (cookies.txt)
 * doesn't keep track of when a cookie was set or if it is foreign.  So, we keep track
 * of this information in our own cookie file, cookiepanel.txt.  This function saves
 * to that file.
 */
	var cpf = getCookiePanelFile();
	var msg = "";
	
	// If the user just installed the CookiePanel, then there will be no
	// cookiepanel.txt file.  In such a case, the CookiePanel will show
	// all cookies as new.  Let the user know that this information won't
	// be accurate for the time being.
	if(!cpf.exists()) {
		msg = "Note about the CookiePanel extension:"
			+ "\n"
			+ "\n"
			+ "\nIt appears that this is the first time you've used the"
			+ "\nCookiePanel extension on this computer.  Thanks for"
			+ "\ngiving it a try."
			+ "\n"
			+ "\n"
			+ "\nYou might notice that the cookies you have now all"
			+ "\nshow up as though you received them today.  In other"
			+ "\nwords, their creation date is today, even though"
			+ "\nmaybe they were actually created before.  That's"
			+ "\nbecause without the CookiePanel installed and open, Firefox"
			+ "\ndoesn't keep track of the creation date of cookies."
			+ "\nSo, for now, the CookiePanel thinks all of your cookies"
			+ "\nare new cookies."
			+ "\n"
			+ "\n"
			+ "\nDon't worry about this now.  Just keep it in mind in"
			+ "\ncase you notice something later."
			+ "\n"
			+ "\n"
			+ "\nIf you have any questions, please don't hesitate to"
			+ "\nemail aquinn@cs.washington.edu.";
			if(ENABLE_USER_ALERTS)
				alert(msg);
			reviewNewCookies();
			setTimeout("reviewNewCookies();",MILLISECONDS_TO_HIGHLIGHT_NEW_COOKIES+100);
	}
	var i = 0;
	var s = "";
	cpf.open("w");
	for( var cookieKeyName in cookiesHash ) {
		s += serializeCookie(cookiesHash[cookieKeyName]);
	}
	cpf.write(s);
	cpf.close();
}

function serializeCookie(cookie) {
/*
 * Serialize a cookie for storage in the cookiepanel.txt file.
 */
	var TAB = "\t";
	var s = cookie.name + TAB
	  + cookie.value + TAB
	  + cookie.isDomain + TAB
	  + cookie.path + TAB
	  + cookie.isSecure + TAB
	  + cookie.expirationDate + TAB
	  + cookie.host + TAB
	  + cookie.isForeign + TAB
	  + cookie.expirationCategory + TAB
	  + cookie.creationDate + TAB
	  + cookie.domain + "\n";
	return s;
}

function unserializeCookie(cookieText) {
/*
 * Unserialize a cookie that was stored in the cookiepanel.txt file.
 */
	var itemNumber = 0;
	var cookieArray = [];
	var i = 0;
	var textLength = cookieText.length;
	var ch = "";
	cookieArray[0] = "";
	cookieArray = cookieText.split("\t");
	var name = cookieArray[0];
	var value = cookieArray[1];
	var isDomain = true;
	if( cookieArray[2]=="false" )
		isDomain = false;
	var path = cookieArray[3];
	var isSecure = true;
	if( cookieArray[4]=="false" )
		isSecure = false;
	var expirationDate = cookieArray[5];
	var host = cookieArray[6];
	var isForeign = true;
	if( cookieArray[7]=="false" )
		isForeign = false;
	var expirationCategory = Number(cookieArray[8]);
	var creationDate = cookieArray[9];
	var domain = cookieArray[10];
	var cookie = createCookie(name, value, isDomain, path, 
		isSecure, expirationDate, host, isForeign, expirationCategory, 
		creationDate); 
	return cookie;
}

function getCookieForKey(key) {
/*
 * Just a quick convenience function to open the cookiepanel.txt file and get the
 * cookie associated with the given key.  There is no error checking to see if the
 * key actually existed.  Since every reference to this function comes from sources
 * that would contain only cookies that are definitely in the cookiepanel.txt file,
 * it should be safe.
 */
	var filedCookiesHash = getFiledCookiesHash();
	return filedCookiesHash[key];
}


/**************************************************************
 *
 * Functions for GETTING INFORMATION FROM COOKIES
 *
 **************************************************************/

function makeCookieKeyName(cookie) {
/*
 * Combine the host name and cookie name to create a unique identifier
 * for the key.  This will be used for storing cookies in hash tables.
 * This key will also be used for the 'value' attribute in the listitems
 * in the listbox.
 */
	var cookieKeyName = "";
	cookieKeyName = cookie.host + "\t" + cookie.name;
	return cookieKeyName;
}

function detectIfCookieIsForeign(host) {
/*
 * Stub code.  TODO:  Write a code here to test the given host and
 * compare it with the host names associated with all open windows.
 * .
 * Update 9/27/05: This appears to be impossible.  It's very possible
 * to come up with a good guess if a cookie is foreign, but there's
 * nothing that is dependably accurate.  If you compare with the currently
 * open window or windows, there's a chance that the page that set the
 * cookie has already been closed and it will be falsely reported as
 * a third party cookie.  Actually, this strategy would, on the whole, do
 * a better job of recognizing ad-related cookies than the built-in
 * Mozilla/Firefox code because many times sites circumvent third-party
 * cookie blocking by doing a quick redirect through the advertiser's domain.
 * If we just compare the host of the cookie with whatever is open at the
 * moment (perhaps comparing with all open windows), that would defeat
 * that tactic.  But, then the user, after setting their browser to block
 * third party cookies, would see the third-party cookies and assume that
 * something was broken.  So, for the time being, we're putting the
 * third party cookie detection on hold.  This was suggested a few years
 * ago in a phone conversation with Steve Morse, a Netscape developer
 * who worked on the cookie features of Mozilla/Netscape.
*/
	return false;
}

function makeCookieDomainDisplay(cookie) {
/*
 * Trim the domain name to be as short as possible for display
 * in the listbox.  We try to trim off the "www" and such and
 * trim off the ".com", ".net", ".edu", ".co.jp", ".ed.uk"
 * and such.  Hopefully, we'll end up with only the part that
 * the user would recognize as a company name.
 */
	var displayDomain;
	if(cookie.host=="" || cookie.host=="localhost" || cookie.host=="127.0.0.1") {
		displayDomain = "(local computer)";
	}
	else {
		displayDomain = cookie.domain.substring(0,cookie.domain.lastIndexOf("."));
	}
	var lastThreeChars = displayDomain.substring(displayDomain.length-3,displayDomain.length);
	if( lastThreeChars==".co" || lastThreeChars==".ne" || lastThreeChars ==".ac" || lastThreeChars==".go" ) {
		displayDomain = displayDomain.substring(0,displayDomain.length-3);
	}
	return displayDomain;
}

function makeCookieItemText(cookie) {
/*
 * Create the text that will be shown in the listbox for the individual
 * cookie item.
 */
	var s = "  ";
	s = s + makeCookieDomainDisplay(cookie)+": "; //add short domain

	// Add a suffix to indicate the cookie's age.
	if( cookie.expirationCategory==EXP_TEMPORARY )
		s = s + "temporary";
	else if( cookie.expirationCategory==EXP_TODAY )
		s = s + "today";
	else if( cookie.expirationCategory==EXP_YEAR )
		s = s + "exp. within 1 year";
	else if( cookie.expirationCategory==EXP_NEVER )
		s = s + "long term";
	else
		s = s + "ERROR";
	
	// Tack on the name of the cookie in parentheses.  If we ever get a cookie
	// with a long name, this will get messy, but it wouldn't cause any major
	// problems as far as I know.
	s = s + " (\"" + cookie.name + "\")";

	// If the cookie is new, then label it as such.
	if( cookieIsNew(cookie) ) {
		s += " <-- NEW";
	}
	return s;
}

function getExpirationCategory(expirationDate) {
/*
 * Look at the expiration date on the cookie.  Based on that,
 * categorize the cookie as temporary, expiring this year, or
 * permanent.  Note that there are no cookies which are actually
 * permanent.  Anything that expires in longer than a year is
 * given the label "permanent" to mean that it will be around for
 * a long time.
 */
	var today = new Date();
	var thisYear = today.getFullYear();
	var thisMonth = today.getMonth();
	var thisDay = today.getDate();
	var tomorrow = new Date(thisYear, thisMonth, thisDay+1);
	var nextYear = new Date(thisYear+1, thisMonth, thisDay+1);
	var cookieExpirationDate = new Date(expirationDate*1000);
	if( expirationDate==0 )
		return EXP_TEMPORARY;
	else if( cookieExpirationDate > nextYear)
		return EXP_NEVER;
	else if( cookieExpirationDate >= tomorrow )
		return EXP_YEAR;
	else
		return EXP_TODAY;
}

function constructLabelForCookie(cookie) {
/*
 * Construct the label that will be used for the subheading in the listbox.
 * (e.g. "Cookies from doubleclick", "Past hour", "Permanent cookies", etc.)
 */
	var group = 0;
	var label = "";
	if( g_sortMode == SORT_MODE_DOMAIN ) {
		label = "Cookies from "+makeCookieDomainDisplay(cookie);
	}
	else if( g_sortMode == SORT_MODE_EXPIRATION ) {
		label = EXP_LABELS[cookie.expirationCategory];
	}
	else if( g_sortMode == SORT_MODE_CREATION ) {
		var cookieAge = (new Date()).getTime() - cookie.creationDate;
		if( cookieAge<3600000 ) {          // 3600000=60*60*1000=number of milliseconds in an hour
			label = CREATION_LABEL_HOUR;
		}
		else if( cookieAge<86400000 ) {    // 86400000=24*60*60*1000=number of milliseconds in a day
			label = CREATION_LABEL_DAY;
		}
		else if( cookieAge<604800000 ) { // 604800000=24*60*60*1000*7=number of milliseconds in a week
			label = CREATION_LABEL_WEEK;
		}
		else if( cookieAge<2592000000 ) {  // 2592000000=24*60*60*1000*30=number of milliseconds in 30 days
			label = CREATION_LABEL_MONTH;
		}
		else if( cookieAge<31536000000 ) { // 31536000000=24*60*60*1000*365=number of milliseconds in 365 days
			label = CREATION_LABEL_YEAR;
		}
		else {
			label = CREATION_LABEL_FOREVER;
		}
	}
	return label;
}

function cookieIsNew(cookie) {
/*
 * Determine if the cookie was set in the last few seconds.  This is generally used for the
 * highlighting of new cookies.
 */
	var result;
	if( ((new Date()).getTime() - cookie.creationDate) < MILLISECONDS_TO_HIGHLIGHT_NEW_COOKIES ) {
		result = true;
	}
	else {
		result = false;
	}
	return result;
}

function getCssClassForCookie( cookie ) {
/*
 * Depending on whether the cookie is new (last few seconds) or old,
 * foreign or from the originating site, we choose a CSS class which
 * will display that information to the user by way of coloring.
 */
	var cssClass = "";
	if( cookie.isForeign ) {
		if( cookieIsNew(cookie) ) {
			cssClass = CSS_CLASS_NEWFOREIGN;
		}
		else {
			cssClass = CSS_CLASS_FOREIGN;
		}
	}
	else {
		if( cookieIsNew(cookie) ) {
			cssClass = CSS_CLASS_NEWNOTFOREIGN;
		}
		else {
			cssClass = CSS_CLASS_NOTFOREIGN;
		}
	}
	return cssClass;
}


/**************************************************************
 *
 * Functions for SORTING COOKIES
 *
 **************************************************************/

function sortCookies() {
/*
 * This will be called when the user clicks the header button on
 * the listbox to change the sort order of the cookies.
 */
	var headRow = document.getElementById("headRowCell");
	if( g_sortMode==SORT_MODE_DOMAIN ) {
		g_sortMode = SORT_MODE_EXPIRATION;
		headRow.setAttribute("label", SORT_LABEL_EXPIRATION);
	}
	else if( g_sortMode==SORT_MODE_EXPIRATION ) {
		g_sortMode = SORT_MODE_CREATION;
		headRow.setAttribute("label", SORT_LABEL_CREATION);
	}
	else {
		g_sortMode = SORT_MODE_DOMAIN;
		headRow.setAttribute("label", SORT_LABEL_DOMAIN);
	}
	buildPanel();
}

function compareCookieTuple(a,b) {
/*
 * Used by the built in array sorting routine to sort the cookies
 * before putting them into the listbox.
 */
	var result = compareLabels(a[0],b[0]);
	var aText = makeCookieItemText(a[1]);
	var bText = makeCookieItemText(b[1]);
	if( result==0 ) {
		var aDomain = a[1].domain.toLowerCase();
		var bDomain = b[1].domain.toLowerCase();
		if( aDomain < bDomain ) result = -1;
		else if( aDomain > bDomain ) result = 1;
		else result = 0;
	}
	if( result==0 ) {
		var aName = a[1].name.toLowerCase();
		var bName = b[1].name.toLowerCase();
		if( aName < bName ) result = -1;
		else if( aName > bName ) result = 1;
		else result = 0;
	}
	return result;
}

function compareLabels(a,b) {
/*
 * Compare the headings that appear in the listbox in bold type.
 * This essentially determines in what order the groups appear.
 * For example, this says that when you're in Sort By Expiration
 * mode, cookies which expire soon should be shown at the top.
 */
	var result = 0;
	if(g_sortMode==SORT_MODE_DOMAIN) {
		if(a<b) result = -1;
		else if(a>b) result = 1;
		else result = 0;
	}
	else if(g_sortMode==SORT_MODE_CREATION) {
		switch(a) {
			case CREATION_LABEL_HOUR:
				switch(b) {
					case CREATION_LABEL_HOUR:
						result = 0;
						break;
					case CREATION_LABEL_DAY:
					case CREATION_LABEL_WEEK:
					case CREATION_LABEL_MONTH:
					case CREATION_LABEL_YEAR:
					case CREATION_LABEL_FOREVER:
						result = -1;
				}
				break;
			case CREATION_LABEL_DAY:
				switch(b) {
					case CREATION_LABEL_HOUR:
						result = 1;
						break;
					case CREATION_LABEL_DAY:
						result = 0;
						break;
					case CREATION_LABEL_WEEK:
					case CREATION_LABEL_MONTH:
					case CREATION_LABEL_YEAR:
					case CREATION_LABEL_FOREVER:
						result = -1;
				}
				break;
			case CREATION_LABEL_WEEK:
				switch(b) {
					case CREATION_LABEL_HOUR:
					case CREATION_LABEL_DAY:
						result = 1;
						break;
					case CREATION_LABEL_WEEK:
						result = 0;
						break;
					case CREATION_LABEL_MONTH:
					case CREATION_LABEL_YEAR:
					case CREATION_LABEL_FOREVER:
						result = -1;
				}
				break;
			case CREATION_LABEL_MONTH:
				switch(b) {
					case CREATION_LABEL_HOUR:
					case CREATION_LABEL_DAY:
					case CREATION_LABEL_WEEK:
						result = 1;
						break;
					case CREATION_LABEL_MONTH:
						result = 0;
						break;
					case CREATION_LABEL_YEAR:
					case CREATION_LABEL_FOREVER:
						result = -1;
				}
				break;
			case CREATION_LABEL_YEAR:
				switch(b) {
					case CREATION_LABEL_DAY:
					case CREATION_LABEL_WEEK:
					case CREATION_LABEL_MONTH:
						result = 1;
						break;
					case CREATION_LABEL_YEAR:
						result = 0;
						break;
					case CREATION_LABEL_FOREVER:
						result = -1;
				}
				break;
			case CREATION_LABEL_FOREVER:
				switch(b) {
					case CREATION_LABEL_DAY:
					case CREATION_LABEL_WEEK:
					case CREATION_LABEL_MONTH:
					case CREATION_LABEL_YEAR:
						result = 1;
						break;
					case CREATION_LABEL_FOREVER:
						result = 0;
				}
				break;
		}
	}
	else if(g_sortMode==SORT_MODE_EXPIRATION) {
		var i=0;
		var aVal=0, bVal=0;
		for(i=EXP_LABELS.length-1; i>=0; i--) {
			if(EXP_LABELS[i]==a)
				aVal = i;
			if(EXP_LABELS[i]==b)
				bVal = i;
		}
		if(aVal<bVal) result = -1;
		else if(aVal>bVal) result = 1;
		else result = 0;
	}
	else {
		// shouldn't get here
		result = -1;
	}
	return result;
}




/**************************************************************
 *
 * Functions for the COOKIE CLASS
 *
 **************************************************************/

function convertCookie( mozillaCookie ) {
/*
 * Convert a cookie object received from the built in cookie interface
 * into an object for use with the CookiePanel.  The CookiePanel
 * cookie object contains a little extra information, currently the
 * creation date and a flag indicating if the cookie is foreign.
 * (As of 9/27/05, we have disabled the detection of foreign cookies
 * but we still pass around a dummy flag, just in case we find a good
 * way to deal with them in the future.)
 */
	var name = mozillaCookie.name;
	var value = mozillaCookie.value;
	var isDomain = mozillaCookie.isDomain;
	var path = mozillaCookie.path;
	var isSecure = mozillaCookie.isSecure;
	var expirationDate = mozillaCookie.expires;
	var host = mozillaCookie.host;
	var isForeign = detectIfCookieIsForeign(host);
	var expirationCategory = getExpirationCategory(expirationDate);
	var creationDate = new Date().getTime(); // Will likely be overridden.
	var cookie = createCookie(name, value, isDomain, path, 
		isSecure, expirationDate, host, isForeign, expirationCategory, creationDate);
	return cookie;
}

function Cookie() {
/*
 * Class declaration for Cookie.  Has no methods and all data fields are added by
 * the createCookie function, so this declaration needs no contents.
 */
}

function createCookie(name, value, isDomain, path, 
		isSecure, expirationDate, host, isForeign, expirationCategory, creationDate) {
/*
 * Constructor for Cookie class.
 */
	var c = new Cookie();

	// information that was in the built-in Mozilla cookie
	c.name = name;
	c.value = value;
	c.isDomain = isDomain;
	c.path = path;
	c.isSecure = isSecure;
	c.expirationDate = expirationDate;
	c.host = host;

	// new information needed for CookiePanel
	c.isForeign = isForeign;
	c.expirationCategory = expirationCategory;
	c.creationDate = creationDate;

	if(typeof(host)=="undefined") {
		dmp("bad cookie: "+printable(c));
		c.domain = "???";
		c.host="???";
	}
	else {
		c.domain = makeCleanDomain(host);
	}
	return c;
}




/**************************************************************
 *
 * Functions for TIMEOUTS AND OBSERVERS
 *
 **************************************************************/

var cookieReloadListener = {
/*
 * Contains the function which is called when a new cookie is set.
 *
 * This was taken partly from CookieViewer.js in the built-in wallet code
 */
	observe: function(subject, topic, state) {
		if( topic==observerString) {
			setTimeout("respondToPanelUpdateTimer()",MILLISECONDS_DELAY_TO_REBUILD_PANEL);
			timersSet++;
			if( state=="cookies" || state=="added" || state=="cleared" || state=="deleted") {
			}
			else {
				admp("caught an unknown cookie state, state==" + state);
			}
		}
	}
}

function respondToPanelUpdateTimer() {
/*
 * When the user opens a page which sets cookies, an event is triggered for every
 * cookie that is set or deleted.  If we updated the panel every time, it would slow
 * down the browser significantly and unacceptably.  So, essentially, we don't update
 * the panel until a second has elapsed since the last cookie event.  Every time an
 * event is caught, we set a timer to call this function AND we increment the timerSet
 * global variable.  When this function is called, we only proceed to buildPanel iff
 * there are no other timers currently pending, per the timersSet global variable.
 */
	timersSet--;
	if( timersSet==0 ) {
		buildPanel();
	}
	else if ( timersSet<0 ) {
		timersSet = 0;
	}
}



/**************************************************************
 *
 * Functions for DELETING AND VIEWING COOKIES
 *
 **************************************************************/

function deleteAllCookies() {
/*
 * Called when the user clicks a button to clear all cookies.
 */
 	if(!panelIsEmpty()) {
		if( confirm("Do you really want to delete all of your cookies?\n\nClick OK to delete them.  Click cancel to leave them.") ) {
			var cookiemanager = Components.classes["@mozilla.org/cookiemanager;1"].getService();
			cookiemanager = cookiemanager.QueryInterface(Components.interfaces.nsICookieManager);
			cookiemanager.removeAll();
			buildPanel();
			cookiemanager = null;
		}
	}
}

function deleteSelectedCookies() {
/*
 * Called when the user selects a cookie(s), right clicks, and chooses delete.
 */
	var cookieListbox = document.getElementById("cookieList");
	var selectedItems = cookieListbox.selectedItems;
	var item = null;
	var cookiemanager = Components.classes["@mozilla.org/cookiemanager;1"].getService();
	cookiemanager = cookiemanager.QueryInterface(Components.interfaces.nsICookieManager);
	for( var i=0; i<selectedItems.length; i++) {
		item = selectedItems[i];
		if( item.getAttribute("class")!=CSS_CLASS_LABEL) {
			var cookieKey = item.getAttribute("value");
			var cookie = getCookieForKey(cookieKey);
			if(cookie) {
				cookiemanager.remove(cookie.host, cookie.name, cookie.path, false);
			}
		}
	}
	timersSet -= 1;
	buildPanel();
	cookiemanager = null;
}

function viewSelectedCookie() {
/*
 * Called when the user right clicks a cookie and selects View.
 *
 * Displays part of the contents of the cookie in a dialog box.  Skips some of the
 * information that would be confusing to the user.  Also, uses slightly different
 * language than the built-in cookie manager, trying to be a little more understandable.
 * Also, includes the creation date, which the built in cookie manager doesn't give.
 */
	var cookieListbox = document.getElementById("cookieList");
	var selectedItems = cookieListbox.selectedItems;
	var item = null;
	var cookiemanager = Components.classes["@mozilla.org/cookiemanager;1"].getService();
	cookiemanager = cookiemanager.QueryInterface(Components.interfaces.nsICookieManager);
	for( var i=0; i<selectedItems.length; i++) {
		item = selectedItems[i];
		if( item.getAttribute("class")!=CSS_CLASS_LABEL && item.current) {
			var cookieKey = selectedItems[i].getAttribute("value");
			var cookie = getCookieForKey(cookieKey);
			if(cookie) {
				var cookieValue = cookie.value;
				if( cookieValue.length>50 ) {
					cookieValue = cookieValue.substring(0,48)+"..."
				}
				msg = "Cookie details:"
						+"\n"
						+"\nCookie's name:  "+cookie.name
						+"\nCookie's contents:  "+cookieValue
						+"\nCreated by a web site at:  "+makeCleanDomain(cookie.host)
						+"\nCreated on or before:  "+(new Date(parseInt(cookie.creationDate))).toLocaleString()
						+"\nSet to remain on your computer until:  "+(new Date(cookie.expirationDate*1000)).toLocaleString();
				alert(msg);
			}
		}
	}
	cookiemanager = null;
}




/**************************************************************
 *
 * Functions related to the COOKIE ACCEPT/DELETE PREFERENCE
 *
 **************************************************************/

function getCookiePreference() {
/*
 * Gets the user preference that says if the user wants to accept all cookies (0),
 * block all cookies (2), or block only third party cookies (1).
 */
	var pref = Components.classes["@mozilla.org/preferences-service;1"]
		.getService(Components.interfaces.nsIPrefBranch);
	var result = pref.getIntPref("network.cookie.cookieBehavior");
	pref = null;
	return result;
}

function setCookiePreference(cookiePreference) {
/*
 * Sets the user preference to say if the user wants to accept all cookies (0),
 * block all cookies (2), or block only third party cookies (1).
 */
	var pref = Components.classes["@mozilla.org/preferences-service;1"]
		.getService(Components.interfaces.nsIPrefBranch);
	pref.setIntPref("network.cookie.cookieBehavior",cookiePreference);
	pref = null;
}

function setCookiePreferenceWidget() {
/*
 * Gets the users preference about accepting/blocking cookies and updates the
 * combo box with that preference.
 *
 * There is a known issue with this functionality.  If the user changes the
 * setting in the built-in preferences dialog box, this widget won't be updated
 * right away.  It will, however, be updated next time there is a update to the
 * CookiePanel.  Really, we need a preference observer or something like that.
 */
	var comboBox = document.getElementById("cookiePanelPrefMenuList");
	var cookiePreference = getCookiePreference();
	comboBox.selectedIndex = cookiePreference;
}


/**************************************************************
 *
 * OTHER FUNCTIONS
 *
 **************************************************************/

function reviewNewCookies() {
/*
 * Run through the whole list and update the highlighting of new cookies.
 * When a cookie first comes in, it is highlighted in yellow and it has
 * a label saying "<-- NEW" attached to it.  After a few seconds, that label
 * and highlighting go away.  This function takes care of that.
 */
	var cookieListbox = document.getElementById("cookieList");
	var cookieListboxItems = cookieListbox.getRowCount();

	// Don't proceed if the list is empty.  Otherwise, you'll get an error
	// when you try to access the first item, which doesn't exist.
	if(cookieListboxItems==0) {
		return;
	}
	var filedCookiesHash = getFiledCookiesHash();
	var item = cookieListbox.getItemAtIndex(0);
	var key = "";
	var cssClass = "";
	var itemText = "";
	var cookie = null;
	var startTime = (new Date()).getTime();
	for( var i=cookieListboxItems; i!=0; i-- ) {
		key = item.getAttribute("value");
		cssClass = item.getAttribute("class");
		if( cssClass!=CSS_CLASS_LABEL && hashHasKey(filedCookiesHash,key) ) {
			cookie = filedCookiesHash[key];

			// Update the CSS class.  If the cookie was considered new before
			// but it is no longer new (longer than a few seconds old), then it
			// will lose the highlighting here.
			cssClass = getCssClassForCookie(cookie);
			item.setAttribute("class",cssClass);

			// Update the text so we can remove the "<-- NEW" label from cookies that are no longer new.
			itemText = makeCookieItemText(cookie);
			item.setAttribute("label",itemText);
		}
		item = cookieListbox.getNextItem(item,1);
	}
}

function setCookieCounter(cookieCount) {
/*
 * Update the value shown in the little counter that shows how many
 * cookies are stored.
 */
	switch(cookieCount) {
		case 0:
			document.getElementById("cookieCount").setAttribute("value", "No cookies");
			break;
		case 1:
			document.getElementById("cookieCount").setAttribute("value", "1 cookie");
			break;
		case 300:
			document.getElementById("cookieCount").setAttribute("value", "300 cookies (the limit)");
			break;
		default:
			document.getElementById("cookieCount").setAttribute("value", cookieCount + " cookies");
	}
}

function jslibCheckInst () {
/*
 * Ensure that the jsLib library is installed.  If it hasn't been installed, give the user
 * an opportunity to download it automatically.
 *
 * This function was taken from the authors of jsLib at http://jslib.mozdev.org/jsLilb_glue.html .
 */
  if (typeof(jslib) != "object") {
    const msg = "CookiePanel requires jsLib to properly operate\n"
              + "Would you like to install jsLib now? (browser restart required)";
    if (confirm(msg)) {
      const url = "https://www.mozdevgroup.com/dropbox/jslib/signed/"
                + "jslib_current_signed.xpi";
      try {
        _content.location = url;
      } catch (e) { open(url, "_blank", "width=1,height=1"); }
    }
    throw "Stopping any further code execution ...";
  }
}

function findAllHtmlDocuments() {
/*
 * NOT USED
 *
 * In an effort to determine if a cookie is from the origination site or if it
 * is foreign, I thought maybe I could compare the domain of the cookie with every
 * window that is currently open.  A long time ago, we had foreign cookie detection
 * implemented by comparing the domain of the cookie with the current window.  That
 * didn't work well because often, by the time we got around to updating the CookiePanel
 * panel, a different window was in the foreground.  This was especially true with popups.
 * You go to www.cnn.com (or whatever).  It sets a cookie and immediately launches a popup.
 * When the CookiePanel updates, it compares the domain of the cookie "cnn.com" to the
 * domain of the currently open window, the popup from "obnoxiousads.com".  It wrongly
 * classifies the cookie from cnn.com as a foreign cookie.  I figured I could improve on this
 * by cycling through all open windows and comparing to all of them.  A few problems exist
 * with this.  First, if the user has chosen to block foreign cookies, this might show
 * some cookies as foreign because those cookies circumvent the foreign cookie blocking.
 * This method would actually identify third party cookies that have tried to circumvent such
 * blocking.  But, it would confuse the user and leave them to think either Firefox/Mozilla is
 * broken or (more likely) the extension is broken.  Second, this would be slow.  You'd
 * have numberOfCookies*numberOfOpenWindows compares.  With 10 open windows and the maximum
 * 300 cookies, you'd have up to 3000 compares.
 *
 * So, we've scrapped the idea for now, but I'm still thinking of picking it up in the future.
 */
	var ifaces = Components.interfaces;
	var mediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(ifaces.nsIWindowMediator);
	var winEnum = mediator.getXULWindowEnumerator(null);
	while (winEnum.hasMoreElements()){
		var win = winEnum.getNext();
		if(!(win instanceof ifaces.nsIXULWindow)) {
			continue;
		}
		var iface = ifaces.nsIDocShellTreeItem;
		var docShellEnum = win.docShell.getDocShellEnumerator(iface.typeAll, iface.ENUMERATE_FORWARDS);
		dmp("Window: (");
		while(docShellEnum.hasMoreElements()) {
			var docShell = docShellEnum.getNext();
			if(docShell instanceof ifaces.nsIDocShell) {
				var domDocument = docShell.contentViewer.DOMDocument;
				if(domDocument instanceof HTMLDocument) {
					dmp("\n\t"+docShell);
					dmp("\t" + domDocument);
					dmp("\t" + domDocument.documentURI);
				}
			}
		}
		dmp("\n)\n");
	}
}

function makeCleanDomain(domain) {
/*
 * Remove the left-most part of the domain name.
 */
	var parts = domain.split(".");
	var cleanDomain = "";
	for(var i=parts.length-1; i>=0; i--) {
		cleanDomain = parts[i] + (cleanDomain==""?"":".") + cleanDomain;
		if( parts[i].length > 3 ) {
			break;
		}
	}
	if(cleanDomain.charAt(0)==".") {
		cleanDomain = cleanDomain.substring(1,cleanDomain.length);
	}
	return cleanDomain;
}

function panelIsEmpty() {
/*
 * Return true iff the listbox is empty.
 */
	var cookieListbox = document.getElementById("cookieList"); //for testing only
	return cookieListbox.getRowCount()==0;
}
